"""PyPICloud downloader tool to bypass PyPICloud and pull directly from S3.

Copyright (c) 2015 CCP Games. Released for use under the MIT license.
"""


from __future__ import print_function

import sys
from collections import defaultdict

from . import get_settings
from . import get_bucket_conn
from . import get_package_version
from . import parse_package


def prefer_wheels(package_releases, package, release=None):
    """Given a list of packages, prefer a single wheel if not overridden.

    Args::

        package_releases: a list of S3 keys for package releases
        package: string package name (used for error reporting)
        release: string release requested or None (used for error reporting)

    Returns:
        a single key if it was possible to reduce to one, or None
    """

    versioned = defaultdict(list)
    for package_release in package_releases:
        package_version = get_package_version(package_release, package)
        versioned[package_version].append(package_release)

    # compare versions with pkg_resources.parse_version, find newest
    ver_order = sorted(versioned)
    packages = versioned[ver_order[-1]]

    wheels = []
    eggs = []
    sources = []
    for pkg in packages:
        if pkg.name.endswith(".whl"):
            wheels.append(pkg)
        elif pkg.name.endswith(".egg"):
            eggs.append(pkg)
        else:
            sources.append(pkg)

    if ("--src" in sys.argv and len(sources) == 1) or len(packages) == 1:
        return sources[0]
    elif (not wheels or "--egg" in sys.argv) and len(eggs) == 1:
        return eggs[0]
    elif len(wheels) == 1:
        return wheels[0]
    else:
        raise SystemExit("Found too many results for {}{}:\n  {}".format(
            package,
            "={}".format(release) if release else "",
            "\n  ".join([key.name for key in packages]),
        ))


def download_package(bucket, package, release=None):
    """Gets the download URL for a package, optionally package+release.

    Args:
        bucket: a connected S3 bucket object to look for package in
        package: string package name without version
        release: string release version to get or None for the latest

    Returns:
        string URL to download the package at release or latest
    """

    # figure out key name from package and release requested and what's
    # available in the bucket...
    package_releases = []
    for key in bucket.get_all_keys():
        if key.name.startswith("{}/".format(package)):
            if release is None or release in key.name.partition("/")[2]:
                package_releases.append(key)

    if len(package_releases) == 1:
        package_key = package_releases[0]
    elif package_releases:
        package_key = prefer_wheels(package_releases, package, release)
    else:
        raise SystemExit("Package {}{} not found".format(
            package,
            "={}".format(release) if release else "",
        ))

    write_key(package_key)


def write_key(key):
    """Writes the key to file or sys.stdout.

    If it can write to a file, it will print the filename to stdout.
    """

    if "--url-only" in sys.argv or "--url" in sys.argv:
        print(key.generate_url(300))  # good for 5 minutes
    else:
        if sys.stdout.isatty():
            # open a file and stream the content into it
            filename = key.name.split("/")[1]
            with open(filename, "wb") as openpackage:
                key.get_contents_to_file(openpackage)
            print(filename)
        else:
            # stdout is being piped/redirected somewhere, write to it directly
            key.get_contents_to_file(sys.stdout)


def main():
    """Main command line entry point for downloading."""

    settings = get_settings(download=True)
    bucket = get_bucket_conn(settings.s3)

    for package in settings.items:
        try:
            download_package(bucket, *parse_package(package))
        except Exception as error:
            print("Error downloading {}: {}".format(package, error),
                  file=sys.stderr)
            break



"""Test the common functions found in the __init__ module."""


import os
import mock
import pytest

import pypicloud_tools


DEFAULT_CONFIG = os.path.join(os.path.expanduser("~"), ".pypirc")


def test_get_bucket_conn():
    """Tests the calls to boto to get our bucket object."""

    mock_config = mock.MagicMock(spec=pypicloud_tools.S3Config)
    mock_boto = mock.Mock()
    mock_bucket = mock.Mock()
    mock_boto.get_bucket = mock.Mock(return_value=mock_bucket)

    with mock.patch.object(pypicloud_tools.boto, "connect_s3", return_value=mock_boto) as patched_boto:
        bucket = pypicloud_tools.get_bucket_conn(mock_config)

    assert bucket == mock_bucket
    patched_boto.assert_called_once_with(mock_config.access, mock_config.secret)
    mock_boto.get_bucket.assert_called_once_with(mock_config.bucket)


def test_settings_from_config(config_file):
    """Ensure that the settings are correctly being read from config."""

    mock_options = mock.Mock()
    del mock_options.acl  # unset acl from command line
    mock_options.config = [config_file]
    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "bucket:test-bucket-name",
            "access:test-access-key",
            "secret:test-secret-key",
            "acl:test-acl",
        ]))

    expected_s3 = pypicloud_tools.S3Config(
        "test-bucket-name", "test-access-key", "test-secret-key", "test-acl"
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://test-server/pypi", "test-username", "test-password"
    )

    s3, pypi = pypicloud_tools.settings_from_config(mock_options)

    assert s3 == expected_s3
    assert pypi == expected_pypi


def test_settings_from_config__no_key(config_file):
    """If there is no [pypicloud] key, we should receive None, None."""

    mock_options = mock.Mock()
    mock_options.config = config_file
    with open(config_file, "w") as openconf:
        openconf.writelines("\n".join([
            "[something]",
            "repository:http://fake.com/pypi",
            "username:joe",
            "password:hunter7",
        ]))

    assert pypicloud_tools.settings_from_config(mock_options) == (None, None)


def test_settings_from_config__cmdline_acl(config_file):
    """Ensure the command line acl overrides whatever is in the config."""

    mock_options = mock.Mock()
    mock_options.acl = ["override-acl"]
    mock_options.config = [config_file]
    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "extravals:are-ignored",
            "bucket:bucket-name",
            "access:access-key",
            "secret:secret-key",
            "acl:test-acl",
        ]))

    expected_s3 = pypicloud_tools.S3Config(
        "bucket-name", "access-key", "secret-key", "override-acl"
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://test-server/pypi", "test-username", "test-password"
    )

    s3, pypi = pypicloud_tools.settings_from_config(mock_options)

    assert s3 == expected_s3
    assert pypi == expected_pypi


def test_settings_from_config__no_acl(config_file):
    """Ensure the acl is None when no acl is configured."""

    mock_options = mock.Mock()
    del mock_options.acl
    mock_options.config = [config_file]
    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "extravals:are-ignored",
            "bucket:bucket-name",
            "access:access-key",
            "secret:secret-key",
        ]))

    expected_s3 = pypicloud_tools.S3Config(
        "bucket-name", "access-key", "secret-key", None,
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://test-server/pypi", "test-username", "test-password"
    )

    s3, pypi = pypicloud_tools.settings_from_config(mock_options)

    assert s3 == expected_s3
    assert pypi == expected_pypi


def test_settings_from_config__read_errors(config_file, capfd):
    """Ensure behaviour if/when there's an error reading the config file."""

    mock_options = mock.Mock()
    mock_options.config = config_file
    with open(config_file, "w") as openconf:
        openconf.write("[malform[ed]]\n**&&&$$$$----\nyes=yes\n")

    assert pypicloud_tools.settings_from_config(mock_options) == (None, None)
    out, err = capfd.readouterr()

    assert "contains parsing errors:" in err
    assert not out


def test_parse_args__list_package():
    """Ensure the parser is setup correctly for listing a package."""

    pypicloud_tools.sys.argv = ["list", "test_package"]
    options, parser = pypicloud_tools.parse_args(listing=True)

    assert isinstance(parser, pypicloud_tools.argparse.ArgumentParser)
    assert isinstance(options, pypicloud_tools.argparse.Namespace)

    expected_options = {
        "packages": ["test_package"],
        "bucket": False,
        "server": False,
        "access": False,
        "secret": False,
        "user": False,
        "password": False,
        "config": DEFAULT_CONFIG,
    }
    assert vars(options) == expected_options
    assert "List package(s) from S3, bypassing PyPICloud" in str(parser)


def test_parse_args__list_all():
    """Ensure the parser is setup correctly for listing all packages."""

    pypicloud_tools.sys.argv = ["list", "--config", "fake.config"]
    options, parser = pypicloud_tools.parse_args(listing=True)

    assert isinstance(parser, pypicloud_tools.argparse.ArgumentParser)
    assert isinstance(options, pypicloud_tools.argparse.Namespace)

    expected_options = {
        "packages": [],
        "bucket": False,
        "server": False,
        "access": False,
        "secret": False,
        "user": False,
        "password": False,
        "config": ["fake.config"],
    }
    assert vars(options) == expected_options


def test_parse_args__upload():
    """Ensure the upload keys are populated in the argument parser."""

    pypicloud_tools.sys.argv = ["upload", "--acl", "fake-acl", "fake_file.tar"]
    options, parser = pypicloud_tools.parse_args(upload=True)

    assert isinstance(parser, pypicloud_tools.argparse.ArgumentParser)
    assert isinstance(options, pypicloud_tools.argparse.Namespace)

    expected_options = {
        "files": ["fake_file.tar"],
        "bucket": False,
        "server": False,
        "access": False,
        "acl": ["fake-acl"],
        "secret": False,
        "user": False,
        "password": False,
        "config": DEFAULT_CONFIG,
    }
    assert vars(options) == expected_options
    assert "Upload package(s) to S3, bypassing PyPICloud" in str(parser)


def test_parse_args__download():
    """Ensure the parser is properly configured for downloads."""

    pypicloud_tools.sys.argv = ["download", "--secret", "abc", "fake_package"]
    options, parser = pypicloud_tools.parse_args(download=True)

    assert isinstance(parser, pypicloud_tools.argparse.ArgumentParser)
    assert isinstance(options, pypicloud_tools.argparse.Namespace)

    expected_options = {
        "packages": ["fake_package"],
        "bucket": False,
        "server": False,
        "access": False,
        "secret": ["abc"],
        "user": False,
        "password": False,
        "config": DEFAULT_CONFIG,
    }
    assert vars(options) == expected_options
    assert "Download package(s) from S3, bypassing PyPICloud" in str(parser)


def test_get_settings__no_args():
    """get_settings() requires one boolean and only one boolean to be True."""

    with pytest.raises(RuntimeError):
        pypicloud_tools.get_settings()
    with pytest.raises(RuntimeError):
        pypicloud_tools.get_settings(True, False, True)


@pytest.mark.parametrize("direction", ("upload", "download", "rehost"))
def test_get_settings__needs_remainders(direction):
    """When not listing, the user must provide a package or filename."""

    pypicloud_tools.sys.argv = [direction]
    with pytest.raises(SystemExit):
        pypicloud_tools.get_settings(download=True)


@pytest.mark.parametrize("direction, acl", [("upload", "fake-acl"), ("download", None)])
def test_get_settings__s3_overrides(direction, acl, config_file):
    """If s3 options are given on the command line, they take precedence."""

    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "bucket:bucket-name",
            "access:access-key",
            "secret:secret-key",
        ]))

    pypicloud_tools.sys.argv = [direction]

    if acl:
        pypicloud_tools.sys.argv.extend(["--acl", acl])

    pypicloud_tools.sys.argv.extend([
        "--bucket",
        "fake-bucket",
        "--access",
        "fake-access",
        "--secret",
        "fake-secret",
        "--config",
        config_file,
        "some_file",
        "some_other_file",
    ])

    settings = pypicloud_tools.get_settings(**{direction: True})
    expected_s3 = pypicloud_tools.S3Config(
        "fake-bucket", "fake-access", "fake-secret", acl,
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://test-server/pypi", "test-username", "test-password"
    )

    assert settings.s3 == expected_s3
    assert settings.pypi == expected_pypi
    assert settings.items == ["some_file", "some_other_file"]


@pytest.mark.parametrize("direction", ("upload", "download", "rehost"))
def test_get_settings__pypi_overrides(direction, config_file):
    """If pypi options are given on the command line, they take precedence."""

    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "bucket:bucket-name",
            "access:access-key",
            "secret:secret-key",
        ]))

    pypicloud_tools.sys.argv = [
        direction,
        "--server",
        "http://some-server/pypi",
        "--user",
        "some-user",
        "--password",
        "some-passwd",
        "--config",
        config_file,
        "some_file",
        "some_other_file",
    ]

    settings = pypicloud_tools.get_settings(**{direction: True})
    expected_s3 = pypicloud_tools.S3Config(
        "bucket-name", "access-key", "secret-key", None,
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://some-server/pypi", "some-user", "some-passwd"
    )

    assert settings.s3 == expected_s3
    assert settings.pypi == expected_pypi
    assert settings.items == ["some_file", "some_other_file"]


def test_get_settings__config_fillins(config_file):
    """Ensure the config file fills in any missing arguments."""

    with open(config_file, "w") as openconf:
        openconf.write("\n".join([
            "[pypicloud]",
            "repository:http://test-server/pypi",
            "username:test-username",
            "password:test-password",
            "bucket:bucket-name",
            "access:access-key",
            "secret:secret-key",
        ]))

    pypicloud_tools.sys.argv = [
        "list",
        "--config",
        config_file,
        "some_file",
        "some_other_file",
    ]
    expected_s3 = pypicloud_tools.S3Config(
        "bucket-name", "access-key", "secret-key", None,
    )
    expected_pypi = pypicloud_tools.PyPIConfig(
        "http://test-server/pypi", "test-username", "test-password"
    )

    settings = pypicloud_tools.get_settings(listing=True)

    assert settings.s3 == expected_s3
    assert settings.pypi == expected_pypi
    assert settings.items == ["some_file", "some_other_file"]


def test_get_settings__no_s3_config(config_file, capfd):
    """If no s3 config is found, ensure get_settings raises an error."""

    pypicloud_tools.sys.argv = ["list", "--config", config_file]
    with pytest.raises(SystemExit):
        pypicloud_tools.get_settings(listing=True)

    out, err = capfd.readouterr()
    assert "ERROR: Could not determine S3 settings." in err
    assert DEFAULT_CONFIG in out  # stdout should be a help message...


@pytest.mark.parametrize(
    "package, pkg_name, pkg_ver",
    [
        ("foo-bar", "foo-bar", None),
        ("foo-bar=0.0.1", "foo-bar", "0.0.1"),
        (
            "foo-bar.baz[stuff]=1.2.beta-FINAL.release.2.tar.gz",
            "foo-bar.baz[stuff]",
            "1.2.beta-FINAL.release.2.tar.gz",
        ),
    ],
    ids=("no version", "simple version", "long exact"),
)
def test_parse_package(package, pkg_name, pkg_ver):
    """Ensure package strings are properly parsed."""

    pkg, ver = pypicloud_tools.parse_package(package)
    assert pkg == pkg_name
    assert ver == pkg_ver


if __name__ == "__main__":
    pytest.main(["-v", "-rx", "--pdb", __file__])



"""Tests to ensure the pypicloud_tools upload functions work as expected."""


import os
import mock
import pytest

import pypicloud_tools
from pypicloud_tools import upload


def test_main(capfd):
    """Mock all calls, ensure the command line entry point flow."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(upload, "get_settings", return_value=settings) as settings_patch:
        with mock.patch.object(upload, "get_bucket_conn", return_value=bucket) as bucket_patch:
            with mock.patch.object(upload, "upload_file") as upload_patch:
                with mock.patch.object(upload, "update_cloud") as cloud_patch:
                    upload.main()

    settings_patch.assert_called_once_with(upload=True)
    bucket_patch.assert_called_once_with(settings.s3)
    cloud_patch.assert_called_once_with(settings.pypi)
    upload_patch.assert_called_once_with("faked", bucket, settings.s3)
    out, err = capfd.readouterr()
    assert not err
    assert "PyPICloud server at {} updated".format(settings.pypi.server) in out


def test_main__buries_error(capfd):
    """PyPICloud should only be updated on success."""

    bucket = mock.Mock()
    settings = mock.Mock()
    settings.items = ["faked"]

    with mock.patch.object(upload, "get_settings", return_value=settings):
        with mock.patch.object(upload, "get_bucket_conn", return_value=bucket):
            with mock.patch.object(upload, "upload_file", side_effect=IOError):
                with mock.patch.object(upload, "update_cloud") as cloud_patch:
                    upload.main()

    out, err = capfd.readouterr()
    assert "Error uploading faked: " in err
    assert not out
    assert not cloud_patch.called


def test_update_cloud():
    """Ensure the proper request calls are used to update the PyPICloud API."""

    mock_get = mock.Mock()
    mock_get.raise_for_status = mock.Mock()
    mock_auth = mock.Mock()
    pypi = pypicloud_tools.PyPIConfig("http://fake/pypi/", "joe", "hunter2")

    with mock.patch.object(upload.requests.auth, "HTTPBasicAuth", return_value=mock_auth) as auth_patch:
        with mock.patch.object(upload.requests, "get", return_value=mock_get) as get_patch:
            assert upload.update_cloud(pypi) == mock_get.ok

    mock_get.raise_for_status.assert_called_once_with()
    auth_patch.assert_called_once_with("joe", "hunter2")
    get_patch.assert_called_once_with("http://fake/admin/rebuild", auth=mock_auth)


def test_upload_file(capfd, config_file):
    """Verify the calls made to upload a file to S3."""

    bucket = mock.Mock()
    bucket_key = mock.Mock()
    bucket.get_key = mock.Mock(return_value=bucket_key)
    mock_multipart = mock.Mock()
    mock_multipart.get_all_parts = mock.Mock(return_value=["one"])
    bucket.initiate_multipart_upload = mock.Mock(return_value=mock_multipart)

    s3_config = mock.Mock()
    file_contents = "some content inside this file..."
    with open(config_file, "w") as openfile:
        openfile.write(file_contents)

    with mock.patch.object(upload, "_upload_chunk") as patched_chunk_uploader:
        upload.upload_file(config_file, bucket, s3_config)

    out, err = capfd.readouterr()

    f_name = os.path.basename(config_file)
    expected_name = "{}/{}".format(f_name[:f_name.index("-")], f_name)
    assert "Uploading {} ...".format(expected_name) in out
    assert "done!" in out
    assert "failed!" not in out
    mock_multipart.complete_upload.assert_called_once_with()
    bucket.get_key.assert_called_once_with(expected_name)
    bucket_key.set_acl.assert_called_once_with(s3_config.acl)
    patched_chunk_uploader.assert_called_once_with(
        bucket,
        mock_multipart.id,   # multipart upload id
        1,                   # 50MB chunk number this is
        config_file,         # file to send
        0,                   # start of file
        len(file_contents),  # to the end of file
    )
    bucket.initiate_multipart_upload.assert_called_once_with(
        expected_name, headers={"Content-Type": "application/octet-stream"}
    )


def test_upload_file__failure(capfd, config_file):
    """Ensure the multipart upload is cancelled on error."""

    expected_name = os.path.basename(config_file)
    if "-" in expected_name:
        expected_name = expected_name[:expected_name.index("-")]
    config_file = "{}-0.0.1.tar.gz".format(config_file)
    expected_name = "/".join([expected_name, os.path.basename(config_file)])
    bucket = mock.Mock()
    bucket_key = mock.Mock()
    bucket.get_key = mock.Mock(return_value=bucket_key)
    mock_multipart = mock.Mock()
    mock_multipart.get_all_parts = mock.Mock(return_value=[])
    bucket.initiate_multipart_upload = mock.Mock(return_value=mock_multipart)

    s3_config = mock.Mock()
    file_contents = "some content inside this file..."
    with open(config_file, "w") as openfile:
        openfile.write(file_contents)

    with mock.patch.object(upload, "_upload_chunk") as patched_chunk_uploader:
        upload.upload_file(config_file, bucket, s3_config)

    out, err = capfd.readouterr()

    assert "Uploading {} ...".format(expected_name) in out
    assert "done!" not in out
    assert "failed! :(" in out
    mock_multipart.cancel_upload.assert_called_once_with()
    patched_chunk_uploader.assert_called_once_with(
        bucket,
        mock_multipart.id,   # multipart upload id
        1,                   # 50MB chunk number this is
        config_file,         # file to send
        0,                   # start of file
        len(file_contents),  # to the end of file
    )
    bucket.initiate_multipart_upload.assert_called_once_with(
        expected_name, headers={"Content-Type": "application/octet-stream"}
    )


def test_upload_chunk(capfd):
    """Ensure the chunk uploader is working as expected."""

    bucket = mock.Mock()
    mp_upload = mock.Mock()
    mp_upload.id = 1234
    bucket.get_all_multipart_uploads = mock.Mock(return_value=[mp_upload])

    with mock.patch.object(upload, "FileChunkIO") as patched_filechunkio:
        upload._upload_chunk(bucket, 1234, 1, "some-fake-file", 0, 50)

    patched_filechunkio.assert_called_once_with("some-fake-file", "rb", offset=0, bytes=50)

    mp_upload.upload_part_from_file.assert_called_once_with(
        fp=patched_filechunkio().__enter__(), part_num=1, cb=upload.print_dot
    )

    out, err = capfd.readouterr()
    assert not err
    assert "1" in out


def test_upload_chunks__errors():
    """It should try up to three additional times to upload each chunk."""

    bucket = mock.Mock()
    mp_upload = mock.Mock()
    mp_upload.id = 1234
    mp_upload.upload_part_from_file = mock.Mock(side_effect=IOError)
    bucket.get_all_multipart_uploads = mock.Mock(return_value=[mp_upload])

    with mock.patch.object(upload, "FileChunkIO") as patched_filechunkio:
        with pytest.raises(IOError):
            upload._upload_chunk(bucket, 1234, 1, "some-fake-file", 0, 50)

    assert bucket.get_all_multipart_uploads.call_count == 4


if __name__ == "__main__":
    pytest.main(["-v", "-rx", "--pdb", __file__])
